/*************************************************************************/
/** Autor  : Netrval Jiri                                               **/
/** Datum  : 6.6.2000                                                   **/
/** Program: Reseni hlavolamu '8' metodou navraceni (backtracing)       **/
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define MAXD 6 /*maximalni hloubka prohledavani*/

typedef int OSMA[9]; /*matice (spise by se hodilo oznaceni vektor, ale je */
                     /* to symbolicke zastoupeni matice, takze matice)*/

typedef struct prvek{
  OSMA matice;
  int rule; /*pravidlo, ktere bylo pouzito na nasledujici prvek*/
  struct prvek *next; /*dalsi prvek*/
} PRVEK;

typedef struct { /*pracovni struktura*/
  OSMA matice;
  int rule;
} DATA;

PRVEK *path; /*seznam ('vetev') matic a pravidel*/

OSMA goal = {1,2,3,8,0,4,7,6,5}; /*cilova matice*/

/*--------------------------------------------------*/

/***** Kontrola na cilovy stav *****/
int Term(DATA data)
{
  int i;

  for (i = 0; i < 9; i++) if (data.matice[i] != goal[i]) return(0);
  return(1);
}

/***** Zjistovani, zda neni prekrocena max. hloubka nebo zda se dana matice jiz nevyskytuje v seznamu *****/
int DeadEnd(DATA data)
{
  int i, h = 1;
  PRVEK *sezn;

  sezn = path->next;   /*nebudem to srovnavat sam se sebou*/
  while (sezn != NULL) {
    h++;
    for (i = 0; i < 9; i++) if (sezn->matice[i] != data.matice[i])
      break; /*konci, najde-li prvni odlisny prvek*/
    if (i == 9) return(1);
      /*dojeli jsme na konec jedne matice bez nalezeni jedineho odlisneho */
      /* prvku, v opacnem pripade jsme z cyklu vyskocili driv (i < 9), */
    sezn = sezn->next; /*zkusime dalsi matici*/
  }
  if (h > MAXD)  return(1); /*jsme prilis hluboko, vratime se*/
  return(0); /*jeste muzeme dal*/
}

/***** Vrati souradnici 'diry' (nuly) v matici *****/
int FindZero(OSMA matice)
{
  int i = 0;

  for (i = 0; i < 9; i++) if (matice[i] == 0) return(i);
  return(-1); /*sem to stejne nedojde, ale prekladac si neda rict*/
}

/***** Zjistuje prvni aplikovatelne pravidlo *****/
int AppRules(DATA data)
{
  int i, r;

  i = FindZero(data.matice);
  r = data.rule; /* -1..3 */  /* -1: novy prvek*/
  r++; /* 0..4 */  /*zkusime dalsi pravidlo*/
  while (1) {
    switch (r) { /*podminky aplikovatelnosti*/
      case 0: if ((i != 0) && (i != 3) && (i != 6)) return(0);
      case 1: if ((i != 0) && (i != 1) && (i != 2)) return(1);
      case 2: if ((i != 2) && (i != 5) && (i != 8)) return(2);
      case 3: if ((i != 6) && (i != 7) && (i != 8)) return(3);
      case 4: return(-1);
    }
    r++;/*zkusime dalsi pravidlo*/
  }
}

/***** zkopirovani matice z pomocne struktury 'data' do seznamu 'path' *****/
void CopyMaticePD(DATA *data)
{
  int i;

  for (i = 0; i < 9; i++)  (*data).matice[i] = path->matice[i];
}

/***** zkopirovani matice ze seznamu 'path' do pomocne struktury 'data' *****/
void CopyMaticeDP(DATA data)
{
  int i;

  for (i = 0; i < 9; i++)  path->matice[i] = data.matice[i];
}

/***** Aplikovani zvoleneho pravidla na danou matici *****/
void AppRule(DATA *data, int r)
{
  int i;

  i = FindZero((*data).matice);
  switch (r) { /*zamena na danych mistech*/
    case 0:  (*data).matice[i] = (*data).matice[i-1];
             (*data).matice[i-1] = 0;  break;
    case 1:  (*data).matice[i] = (*data).matice[i-3];
             (*data).matice[i-3] = 0;  break;
    case 2:  (*data).matice[i] = (*data).matice[i+1];
             (*data).matice[i+1] = 0;  break;
    case 3:  (*data).matice[i] = (*data).matice[i+3];
             (*data).matice[i+3] = 0;  break;
  }
}

/***** Srdce a duse programu *****/
void BackTrack(DATA *data)
{
  int r, i;
  PRVEK *pom;

  if ((*data).rule == -1) { /* je to novy prvek*/
    if (Term(*data) == 1) { /*nalezeno reseni*/
      while (path != NULL) { /*uklidime po sobe*/
        pom = path;  path = path->next;  free((void *) pom);
      }
      putchar('\n');
      for (i = 0; i <= 8; i++) printf("%d ", (*data).matice[i]);
      printf(" - nalezene reseni\n");
      return; /*nalezeno reseni*/
    }
    if (DeadEnd(*data) == 1) { /*tato matice uz existuje, nebo jsme hluboko*/
      putchar('\n');
      for (i = 0; i <= 8; i++) printf("%d ", (*data).matice[i]);
      pom = path;  path = path->next;  free((void *) pom); /*navrat o jednu*/
      CopyMaticePD(&(*data)); /*aktualizace pracovni struktury*/
        (*data).rule = path->rule;
        if (path == NULL) { /*byl to koren, nenalezli jsme reseni*/
          printf("\nNenalezeno reseni.\n");
          return; /*neni kam se vratit, konec*/
        }
     }
  }
  putchar('\n');
  for (i = 0; i <= 8; i++) printf("%d ", (*data).matice[i]);
  r = AppRules(*data); /*najdeme vhodne pravidlo*/
  (*data).rule = r; /*ulozime*/
  printf(" %d",(*data).rule);/*vytiskneme*/

  switch (r) {
    case -1: /*nelze aplikovat dalsi pravidlo, vratime se*/
             pom = path;  path = path->next;  free((void *) pom);
             if (path == NULL) {
               printf("\nNenalezeno reseni.\n");
               return; /*neni kam se vratit, konec*/
             }
             CopyMaticePD(&(*data));
             (*data).rule = path->rule;
             break;
    case 0: /*pouzijeme*/
    case 1:
    case 2:
    case 3: if ((pom = (PRVEK *) malloc(sizeof(PRVEK))) == NULL) {
              printf("\nNedostatek pameti!\n\007");
              while (path != NULL) { /*uklidime po sobe*/
                pom = path;  path = path->next;  free((void *) pom);
              }
              exit(1);
            }
            path->rule = r; /*ulozime*/
            pom->next = path;  path = pom; /*pripojime na celo seznamu*/
            AppRule(&(*data), r); /*pouzijeme vybrane pravidlo*/
            CopyMaticeDP(*data); /*ulozime do seznamu*/
            (*data).rule = -1; /*je to novy prvek*/
            path->rule = -1;
            break;
  }
  BackTrack(&(*data)); /*rekurze*/
  return;
}

/***** Hlavni program *****/
int main()
{
  DATA data;

  data.matice[0] = 1;  data.matice[1] = 2;  data.matice[2] = 3;
  data.matice[3] = 0;  data.matice[4] = 6;  data.matice[5] = 4;
  data.matice[6] = 8;  data.matice[7] = 7;  data.matice[8] = 5;
  data.rule = -1; /*inicializovani pocatecni matice */
  printf("\n\n");
  if ((path = (PRVEK *) malloc(sizeof(PRVEK))) == NULL) {
    printf("\nNedostatek pameti!\n\007");
    return(1);
  }
  CopyMaticeDP(data);
  path->rule = data.rule;
  path->next = NULL;
  BackTrack(&data); /*spusteni*/
  return(0);
}
